#!/usr/bin/perl
use lib '/home/martin/ethereal/mgmt';

################################################################################
# Created       : Martin Foster
# Modified      : 17-Oct-2005
################################################################################
#
# Command line - Command line controls as part of Ethereal Realms
# Copyright (C) 2000-2006  Martin Foster

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 29-26503 TWP RD 511
#                 Spruce Grove, Alberta
#                 T7Y 1G4
#
#################################################################################

use CGI qw(:standard);						# Common Gateway Interface
use Number::Format qw(:subs);					# Number formatting library
use XML::RSS;							# RSS feed generation
use Socket;							# Socket library
use strict;							# Strict variable enforcement

use Ethereal::Database;						# Database handler
use Ethereal::Home;						# Home page handler
use Ethereal::Mail;						# Mail handler
use Ethereal::Param;						# Parameter handling
use Ethereal::Pull;						# Post pulling
use Ethereal::Rotate;						# Rotate library
use Ethereal::Weather;						# Weather forcasts

################################################################################
# Data Members
################################################################################
my $site;							# Site schema	
my $func;							# Function command
my $type;							# Type to expire

my $count;							# Counter
my $database;							# Database handle

my @sites;							# List of sites

################################################################################
# Program Area
################################################################################

	# Handling of defaults
	# Domain lacking
	if ((defined($ARGV[0]))
	 && ($ARGV[0] =~ /^-\w+/))
	{
		# Force defaults
		$site = 'all';

		# Pull the rest
		$func = shift(@ARGV);
		$type = shift(@ARGV);
	}

	# Domain specified
	else
	{
		# Pull the rest
		$site = shift(@ARGV);
		$func = shift(@ARGV);
		$type = shift(@ARGV);
	}

	# Handle process
	# Only continue if parameters are sent
	if (defined($func))
	{

		# Connect to Database
		$database = new Ethereal::Database();

		# Disabled shared memory
		$database->{'ADMN'} = 'true';
		
		# Connect to database
		$database->Connect();


		# Population of array
		# Global listing
		if (($site eq 'all')
		 && ($database->{'MULTI'} != 0))
		{
			# Pull list of active sites
			$database->GetList(\@sites, "SELECT SiteSchema
				FROM public.Search
				GROUP BY SiteSchema");
		}

		# Unique site only
		else
		{
			# Populate hash
			push(@sites, $site);
		}

		# Loop as necessary
		foreach $site (@sites)
		{

			# Increment
			$count++;
			
			# Set search paths if necessary
			$database->Write("SET SEARCH_PATH TO $site,public") 
			 if ($database->{'MULTI'} != 0);

			
			# Function selections	
			# Import
			if ($func eq '-import')
			{ ImportFortunes($database, $type); }

			# Posts
			if ($func eq '-post')
			{ PurgePosts($database, $type); }

			# Display banner
			if ($func eq '-banner')
			{ PrintBanner($database, $type); }

			# End all streams
			if ($func eq '-stream')
			{ PrintStream($database, $type); }

			# End all streams
			if ($func eq '-notice')
			{ PrintNotice($database, $type); }

			# List ban
			if ($func eq '-ban')
			{ ListBan($database); }

			# List crypt
			if ($func eq '-crypt')
			{ ListCrypt($database); }

			# Listing of Emails
			if ($func eq '-list')
			{ ListEmail($database, $type); }

			# Realm stats
			if ($func eq '-realm')
			{ ListRealms($database); }

			# Cumulative statistics reset
			if ($func eq '-reset')
			{ ResetStats($database); }

			# Listing useage summary
			if ($func eq '-summary')
			{ ListSummary($database, $type); }

			# Forcast update
			if ($func eq '-forcast')
			{ ForcastUpdate($database); }

			# Purge parameters
			if ($func eq '-purge')
			{ PurgeParam($database); }
			
			# RSS feed generation
			if ($func eq '-rss')
			{ GenRSS($database, $type); }
	

			# Purge parameters
			if ($func eq '-vacuum')
			{ Vacuum($database, $type); }
	
			# Expiery control
			if ($func eq '-expire')
			{
				# Specific expire
				if (defined($type))
				{
					# Banings
					if ($type eq 'ban')
					{ ExpireBans($database); }

					# Remove old internal mail
					if ($type eq 'mail')
					{ ExpireMail($database); }

					# Offline posts
					elsif ($type eq 'offline')
					{ ExpireOffline($database); }

					# Puppeteers
					elsif ($type eq 'puppeteer')
					{ ExpirePuppeteer($database); }

					# Puppets
					elsif ($type eq 'puppet')
					{ ExpirePuppet($database); }

					# Private realms
					elsif ($type eq 'realm')
					{ ExpireRealm($database); }

					# Registrations
					elsif ($type eq 'registration')
					{ ExpireRegistrations($database); }
				}
	
				# General expire
				else
				{
					# Run though all methods
					ExpirePuppeteer($database);
					ExpireRegistrations($database);
					ExpireOffline($database);
					ExpireRealm($database);
					ExpireBans($database);
					ExpireAudit($database);
					ExpireMail($database);
					ExpirePuppet($database);
					
					# Cleanup
					Vacuum($database);
				}
			}
		}
	}

	# Display help
	unless (defined($func))
	{
		print "Ethereal Realms command-line interface:\n";
		print "Usage: erealms.pl {all|schema} -{function} [additional]\n\n";
		print "-ban\n";
		print "\tSummary of banned users\n\n";
		print "-banner\n";
		print "\tPost banner to the active realms\n\n";
		print "-crypt\n";
		print "\tProduces output for Apache password files\n\n";
		print "\-import\n";
		print "\tImports BSD fortune files into the database\n\n";
		print "-list\n";
		print "\tDisplay a list of all E-Mail addresses of members\n\n";
		print "-forcast\n";
		print "\tUpdates the forcast information for all realms making use of it\n\n";
		print "-post [DAYS]\n";
		print "\tRemoves old posts from the table.\n\n";
		print "-purge\n";
		print "\tPurges the parameter list\n\n";
		print "-realm\n";
		print "\tSummary for a realms specifically\n\n";
		print "-reset\n";
		print "\tResets cumulative statistics\n\n";
		print "-stream\\-notice [MESSAGE]\n";
		print "\tTerminate all active streams \\ Simply send a notice\n\n";
		print "-summary\n";
		print "\tSummary of room useage\n\n";
		print "-rss\n";
		print "\tGenerates RSS feeds for various aspects of the site\n";
		print "-vacuum {complete|fast|verbose}\n";
		print "\tCleans tables for the system\n\n";
		print "-expire\n";
		print "\tWill remove expired entires.  When called alone all are removed\n";
		print "\tor if specified only a specific task will take place as listed\n";
		print "\tbelow:\n";
		print "\t\taudit        - Remove old audit entries\n";
		print "\t\tban          - Remove old bannings\n";
		print "\t\tmail         - Remove old mail\n";
		print "\t\toffline      - Remove offline posts\n";
		print "\t\tpuppet       - Remove inactive puppets\n";
		print "\t\tpuppeteer    - Remove inactive puppeteers\n";
		print "\t\trealm        - Remove inactive private realms\n";
		print "\t\tregistration - Remove old incomplete registrations\n\n";
	}

################################################################################
# Sub-Routines
################################################################################

####################
# Display Hostname
#
# Will always return the hostname, instead of allowing a IP address to peer
# through.  Useful for adding information to certain reports.

sub DisplayHostname
{
	####################
	# Data members

	my $addr = shift;					# Address or hostname

	my $host;						# Hostname
	my $pked;						# Packed binary address


	####################
	# Program area

	# Generate packed address
	$pked = inet_aton($addr);

	# Reverse lookup address
	$host = gethostbyaddr($pked, AF_INET);

	# Make sure values are defined
	$host = (defined($host)) ? $host : '~~~';


	# Return host address
	return $host;
}


####################
# Format Numbers
#
# Will make use of Number::Format to correctly format numbers using decimal
# but will also add in spacing as to avoid problems with tabulations under
# certain cases.
#
sub FormatNum
{
	####################
	# Data members

	my $num   = shift;					# Number to format
	my $limit = shift;					# Number of spaces

	my $count;						# Actual characters used


	####################
	# Program area

	$limit = (defined($limit)) ? $limit : 10;

	# Format
	$num = format_number($num, 0, 0);

	# Fetch length and limit
	$count = length($num);
	$count = $limit - $count;

	# Return formatted
	return (' ' x $count . $num . ' ');
}

####################
# Generate RSS Feeds
#
# Used to generate certain RSS feeds used on the site, such as
# being able to see the amount of people in public realms
# new works to display et cetera.

sub GenRSS
{
	####################
	# Data members
	my $database = shift;					# Database handle
	my $func     = shift;					# Function

	my $cgi;						# CGI handle
	my $param;						# Parameter handler
	my $rss;						# RSS feed handle
	
	my $res;						# Results hash
	my $statement;						# Database hash

	my $date;						# RSS formatted date
	my $descr;						# Description
	my $loc;						# Location of script
	my $realm;						# Singular realm
	my $url;						# Generated link
	my $title;						# Generated title
	
	my %gal;						# Gallery hash
	my %portal;						# Portal hash
	my %system;						# System hash

	####################
	# Program area

	# Link with hashe
	$database->GetHashSystem(\%system);

	# Fetch date
	($date) = $database->DataGetDateRSS();
	
	# Portal listing
	if ($func eq 'portal')
	{
		# Create instance
		$cgi = new CGI;
		$rss = new XML::RSS (version => '1.0');
				
		# Initialize header
		$rss->channel(
			title       => $system{'SetInfoServer'},
			link        => $system{'SetUrl'},
			description => $system{'SetDescription'},
			dc => {
				creator    => $system{'SetInfoContactAddress'},
				publisher  => $system{'SetInfoContactAddress'},
				date       => $date,
				language   => 'en-us',
  			 }
		 );

		# Link wish hash
		$database->GetHashPortal(\%portal);

		
		# Sets location
		$loc = $system{'LocRSSPortal'};

		# Format and initialize
		$database->Pull(\$statement, "SELECT
			  RealmName AS \"Realm\",
			  COUNT(*)  AS \"Count\"
			FROM Portal
			GROUP BY RealmName
			ORDER BY RealmName"
		 );

		# Loop and handle
		while ($res = $statement->fetchrow_hashref())
		{
			# Generate URL
			$url = $cgi->escape($res->{'Realm'});
			$url = $portal{'LnkQuickLink'} . "/" . $url;
			
			# Insert entry
			$rss->add_item(
				title       => $res->{'Realm'} . ' (' . $res->{'Count'} . ')',
				link        => $url
			 );
		}
	}

	# Listed new works
	elsif ($func eq 'works')
	{
		# Link wish hash
		$database->GetHashGallery(\%gal);

		# Create instance
		$rss = new XML::RSS (version => '1.0');
				
		# Initialize header
		$rss->channel(
			title       => $system{'SetInfoServer'},
			link        => $system{'SetUrl'},
			description => $system{'SetDescription'},
			dc => {
				creator    => $system{'SetInfoContactAddress'},
				publisher  => $system{'SetInfoContactAddress'},
				date       => $date,
				language   => 'en-us',
 			 },
			syn => {
				updatePeriod     => 'hourly',
				updateFrequency  => '1',
				updateBase       => '1901-01-01T00:00+00:00'
			 },
		 );

		# Sets location
		$loc = $system{'LocRSSWorks'};


		# Generate and initialize
		$database->Pull(\$statement, "SELECT
			 w.WorkName        AS \"WorkName\",
			 w.WorkID          AS \"WorkID\",
			 w.WorkDescription AS \"WorkDescription\",
			 g.GalleryID       AS \"GalleryID\",
			 g.GalleryType     AS \"GalleryType\",
			 g.GalleryPenName  AS \"GalleryPenName\",
			 getDateRSS(w.WorkTimestamp) AS \"WorkTimestamp\"
			FROM Gallery g, Work w
			WHERE g.GalleryID=w.GalleryID
			  AND g.GalleryPrivacy='no'
			  AND g.GalleryType IN ('image','text')
			ORDER BY w.WorkTimestamp DESC LIMIT 50"
		 );

		# Loop and add in
		while ($res = $statement->fetchrow_hashref())
		{
			# Generate URL, title and description
			$url   = $gal{'SetScript'} . $gal{'LnkGallery'} . '/view/' . $res->{'GalleryID'} . '/' . $res->{'WorkID'};
			$title = ucfirst($res->{'GalleryType'}) . ' - ' . $res->{'WorkName'} . ' by ' . $res->{'GalleryPenName'};
			$descr = (defined($res->{'WorkDescription'})) ? $res->{'WorkDescription'} : 'None';

			# Insert basic item
			$rss->add_item(
				title       => $title,
				link        => $url,
				description => $descr,
				dc => {
					subject  => uc($res->{'GalleryType'}),
					creator  => $res->{'GalleryPenname'},
					created  => $res->{'WorkTimestamp'}
				 }
			 );
		}
	}

	# Who is on
	elsif ($type eq 'who')
	{
		# Link wish hash
		$database->GetHashPortal(\%portal);

		# Create instances
		$cgi   = new CGI;
		$param = new Ethereal::Param($database, $cgi);
		$rss   = new XML::RSS (version => '1.0');
				
		# Initialize header
		$rss->channel(
			title       => $system{'SetInfoServer'},
			link        => $system{'SetUrl'},
			description => $system{'SetDescription'},
			dc => {
				creator    => $system{'SetInfoContactAddress'},
				publisher  => $system{'SetInfoContactAddress'},
				date       => $date,
				language   => 'en-us'
 			 }
		 );

		
		# Pull starting point
		my ($id) = $database->DataGetPostID();

		# Prepare and execute
		$database->Pull(\$statement, "SELECT
			  Po.PuppetName AS \"PuppetName\",
			  Po.RealmName  AS \"RealmName\"
			FROM Post Po, Realm Re
			WHERE Po.RealmName = Re.RealmName
			  AND Re.RealmPublic='yes'
			  AND Po.PostIDNumber > ?::INT
			  AND Po.PuppetName IS NOT NULL
			GROUP BY Po.RealmName, Po.PuppetName
			ORDER BY Po.RealmName, Po.PuppetName", $id);


		# Loop and run through
		while ($res = $statement->fetchrow_hashref())
		{
			# Set formatted realm name
			$realm =  lc($res->{'RealmName'});
			$realm =~ s/\W//gs;
			
			# Location setting
			$loc = $realm unless (defined($loc));


			# New realm
			if ($realm ne $loc)
			{
				# Save up
				$rss->save($system{'LocRSSWho'} . '/' . $loc . '.rss');


				# Start anew
				# Initialize instance
				$rss   = new XML::RSS (version => '1.0');
				
				# Initialize header
				$rss->channel(
					title       => $system{'SetInfoServer'},
					link        => $system{'SetUrl'},
					description => $system{'SetDescription'},
					dc => {
						creator    => $system{'SetInfoContactAddress'},
						publisher  => $system{'SetInfoContactAddress'},
						date       => $date,
						language   => 'en-us'
 					 }
				 );

				# Set new location
				$loc = $realm;
			}

			# Create entry
			$rss->add_item(
				title       => $res->{'PuppetName'},
				link        => $system{'SetUrl'} . $portal{'LnkBio'} . $param->Crypt(BCHAR=>$res->{'PuppetName'}, BPOOL=>'system')
			 );
		}

		# Cleanup
		$loc = $system{'LocRSSWho'} . '/' . $loc . '.rss' if (defined($loc)); 
	}

	# Finish up
	$statement->finish() if (defined($statement));		

	# Save up
	$rss->save($loc) if (defined($loc));
}

####################
# Import Fortunes
#
# Will handle the import of BSD style fortune files into the 'Fortune' table
# in the database.  This in turn will allow for them to be used in later
# retreival.

sub ImportFortunes
{
	####################
	# Data members

	my $database = shift;					# Database handle
	my $idnt     = shift;					# Elapsed days

	my $curr;						# Current line
	my $mess;						# Captured message


	####################
	# Program area

	# Evaluate for speed
	eval
	{
		# Begin the import process
		if (defined($idnt))
		{
			# Loop through lines
			while ($curr = <STDIN>)
			{
				# End of message
				if ($curr =~ /^%/)
				{
					# Re-order spacing
					$mess =~ s/\r//sg;
					$mess =~ s/^\s+/&nbsp;/mg;
					$mess =~ s/(\n)/<BR>$1/sg;
					$mess =~ s/<BR>$1$//s;


					# Display to STDOUT
					print "$mess\n%\n";

					# Update database
					$database->Write("INSERT INTO Fortune
						(FortuneType,
						 FortuneMsg)
						VALUES (?,?)",

					 $idnt,
					 $mess);


					# Reset message
					$mess = undef;
				}


				# Combine lines
				else
				{
					# Clean operation
					$mess = (defined($mess)) ? "$mess$curr" : "$curr";
				}
			}

			# Clear buffer
			if (defined($mess))
			{
				# Re-order spacing
				$mess =~ s/\r//sg;
				$mess =~ s/^\s+/&nbsp;/mg;
				$mess =~ s/(\n)/<BR>$1/sg;
				$mess =~ s/<BR>$1$//s;


				# Display to STDOUT
				print "$mess\n%\n";

				# Update database
				$database->Write("INSERT INTO Fortune
					(FortuneType,
					 FortuneMsg)
					VALUES (?,?)",
	
				 $idnt,
				 $mess);


				# Reset message
				$mess = undef;
			}
		}

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");

		# Undo incomplete changes
		$database->Rollback();
	}
}


####################
# List Ban
#
# Creates a summary view of banned users, for use by administrators
# who may need this information in day to day use. 

sub ListBan
{
	####################
	# Data Members
	my $database = shift;					# Database handle

	my $send;						# Sendmail handle
	my $tstamp;						# Timestamp

	my $query;						# Database query
	my $statement;						# Database statement

	my $addr;						# Singular address
	my $host;						# Singular hostname
	my $spacer;						# Misc. spacer

	my $aliases;						# Singular alias
	my $ban;						# Singular ban
	my $email;						# Singular Email
	my $hostnames;						# List of hostnames
	my $puppets;						# Singular puppet

	my @aliases;						# Get a list of aliases
	my @banned;						# List of banned users
	my @email;						# Email addresses
	my @hostnames;						# List of hostnaMes
	my @puppets;						# List of puppets

	my %system;						# Handle to system


	####################
	# Program area

# Format for the output
my $TOP = '
    Login Name             Known Aliases
    --------------------   ------------------------------
    @<<<<<<<<<<            ^<<<<<<<<<<';

my $MID = '
~~                         ^<<<<<<<<<<';

my $LOW = '
~~  ^<<<<<<<<<<<<<<<<<<<<  ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<';

my $CLUSTER = '
    @<<<<<<<<<<<<<<        ^*';


	# Retreive hash
	$database->GetHashSystem(\%system);


	# Retreive Email
	$database->GetList(\@email, "SELECT PuppeteerEmail
		FROM  Puppeteer
		WHERE PuppeteerAdminPrivs='yes'
		GROUP By PuppeteerEmail");

	# Generate
	($tstamp) = $database->DataGetDate();


	# Initiate Email composition
	# Create instance
	$send = tie(*MAIL, 'Ethereal::Mail', $database);

	# Initial settings
	$send->SetSubject("Status of Banned Users ($tstamp)");

	# Recipients
	$send->AddTo(@email);


	# Banned user summary
	print MAIL 	"   -------------------------------\n",
		 	"   - Banned User Summary         -\n",
			"   -------------------------------\n\n"; 

	# Fetch master userlist
	$database->GetList(\@banned, "SELECT PuppeteerLogin
		FROM Puppeteer
		WHERE PuppeteerBan=?
		  AND PuppeteerAlias IS NULL
		ORDER BY PuppeteerLogin",  'yes');
	

	# Cycle through
	foreach $ban (@banned) 
	{
		# Get list
		$database->GetList(\@aliases, "SELECT PuppeteerLogin
			FROM Puppeteer
			WHERE PuppeteerAlias=?
			ORDER BY PuppeteerLogin", $ban);

		# Append login
		unshift(@aliases, $ban);


		# Hostnames
		$database->GetList(\@hostnames, "SELECT PuppeteerHostname
			FROM  Audit
			WHERE PuppeteerLogin IN (SELECT PuppeteerLogin
				FROM Puppeteer
				WHERE PuppeteerAlias=?
				  OR PuppeteerLogin=?)
			GROUP BY PuppeteerHostname
			ORDER BY PuppeteerHostname", $ban, $ban);
  
		# Puppets
		$database->GetList(\@puppets, "SELECT PuppetName
			FROM Puppet
			WHERE PuppeteerLogin IN (SELECT PuppeteerLogin
				FROM Puppeteer
				WHERE PuppeteerAlias=?
				  OR PuppeteerLogin=?)
			ORDER BY PuppetName", $ban, $ban);


		# Top header
		$send->Write($TOP, $ban, $aliases[0]);

		# Pull one off
		shift(@aliases);


		# Join statements
		$aliases   = join("\n\r", @aliases);
	 	$hostnames = join("\n\r", @hostnames); 
		$puppets   = join("\n\r", @puppets);

		# Reset arrays
		@aliases   = [];
		@hostnames = [];
		@puppets   = [];


		# Mid section
		$send->Write($MID, $aliases);

		# Bottom section
		# Faked headers
		print MAIL "\n    Hostnames              Handle Names\n"; 
		print MAIL "    --------------------   ------------------------------";

		# Actual formatting
		$send->Write($LOW, $hostnames, $puppets);

		# Faked bottom
   		print MAIL "\n   =======================================================\n";
	}


	# General banned IP search pool
	# Hostnames
	$database->GetList(\@hostnames, "SELECT PuppeteerHostname
		FROM  Audit
		WHERE PuppeteerLogin IN (SELECT PuppeteerLogin
			FROM Puppeteer
			WHERE PuppeteerBan='yes'
			ORDER BY PuppeteerLogin)
		GROUP BY PuppeteerHostname
		ORDER BY PuppeteerHostname");


	# Address cluster title
	print MAIL 	"   -------------------------------\n",
		 	"   - Address Cluster             -\n",
			"   -------------------------------\n"; 

	# Information display
	while ($addr = shift(@hostnames))
	{
		# Retreive Hostname
		$host = DisplayHostname($addr);
		$host = lc($host);

		# Write to message
		$send->Write($CLUSTER, $addr, $host);
	} 	 

	# Close and send
	close(MAIL);
}



####################
# List Crypt
#
# Lists the encrypted usernames and passwords for supervisors and the
# administrators to allow for building password files for Apache.

sub ListCrypt
{
	####################
	# Data Members
	my $database = shift;					# Database handle

	my $crypt;						# Action completed
	my $statement;						# Database statement

	
	####################
	# Program area

	# Prepare and execute
	$database->Pull(\$statement, "SELECT 
			  PuppeteerLogin    AS \"PuppeteerLogin\",
			  PuppeteerPassword AS \"PuppeteerPassword\"
			FROM  Puppeteer
			WHERE PuppeteerAdminPrivs='yes'");

	# Crypted Password
	while ($crypt = $statement->fetchrow_hashref())
	{
		print "$crypt->{PuppeteerLogin}:$crypt->{PuppeteerPassword}\n";
	}

	# Finsih call
	$statement->finish();
}


####################
# List EMail
#
# Is used to simply list of all E-Mail addresses of those who are members of the
# chat server.  This is useful to embed within Majordomo or other such mailing
# list system.

sub ListEmail
{
	####################
	# Data members
	my $database = shift;					# Database handle
	my $func     = shift;					# Function

	my $query;						# Query to run
	my $puppeteers;						# Single puppeteer EMail
	my @puppeteers;						# List of all EMails


	####################
	# Program area

	# Make sure of definition
	$func = (defined($func)) ? $func : '-e';

	# Determin query (default, admin, supervisory)
        $query = "SELECT PuppeteerEmail
		  FROM Puppeteer
		  WHERE PuppeteerBan   <> 'yes'
		    AND PuppeteerLogin <> 'system'";
	$query = ($func =~ /^-a/i) ? "$query AND PuppeteerAdminPrivs='yes'" : $query;	
	$query = ($func =~ /^-s/i) ? "$query AND PuppeteerSuperPrivs='yes'" : $query;
	$query = ($func =~ /^-v/i) ? "$query AND PuppeteerVerified='yes'"   : $query;
	
	# Make sure values are unique
	$query = "$query GROUP BY PuppeteerEmail";


	# Retreive puppeteers E_Mail addresses
	$database->GetList(\@puppeteers, $query);

	# Print each E-Mail address
	foreach $puppeteers (@puppeteers)
	{ print "$puppeteers\n"; }
}


####################
# Realm specific summary
#
# Realm specific summaries that will be mailed specifically to the hosts
# that are entrusted with the realms themselves.  This keeps them from
# comparing with other realms in service.

sub ListRealms
{
	####################
	# Data members
	my $database = shift;					# Database handle

	my $pull     = new Ethereal::Pull();			# Pull handle
	my $send;						# Mail handle
	my $res;						# Results handle
	my $statement;						# Query handle

	my $time;						# Selected time
	my $elapsed;						# Elapsed time

	my $end;						# End of search span
	my $start;						# Begin of search span

	my $lhits;						# Local hits in a realm
	my $lposts;						# Local posts in a realm
	my $ghits;						# General hits in a realm
	my $gposts;						# General posts in a realm
	my $ohits;						# Old hits in a realm
	my $oposts;						# Old posts in a realm

	my $count;						# Counter
	my $chars;						# Change parade inventory
	my $host;						# Individual host
	my $pers;						# Personel on parade
	my $realm;						# Individual realm
	my $who;						# Individual person
	my $valid;						# Validity of entry

	my @hosts;						# List of hosts
	my @realms;						# List of realms
	my @whos;						# Who was on
	my @users;						# User list

	my %system;						# System hash


	####################
	# Program area

	# Data retreival
	# Link with hash
	$database->GetHashSystem(\%system);

	# Retreive realms
	$database->GetList(\@realms, "SELECT RealmName 
		FROM RealmSupervisor 
		GROUP BY RealmName
		ORDER BY RealmName");



	# Run through list
	foreach $realm (@realms)
	{
		# Time based statistics
		($start) = $database->DataGet("SELECT getDate(getInterval_Hour('1 day'))");
		($end)   = $database->DataGet("SELECT getDate(getInterval_Hour('0 day'))");


		# Data Retreival
		# Reset to zero
		$ghits  = 0;
		$gposts = 0;

		# List values
		# Retreive userlist
		$database->GetList(\@users, "SELECT Po.PuppeteerLogin || ' (' || Po.PuppetName || ') ' || getWarning(?, Po.PuppeteerLogin)
			FROM  Post Po, 
			      Puppet Ch 
			WHERE Po.PuppeteerLogin IS NOT NULL
			  AND Po.PuppetName IS NOT NULL
			  AND Po.PuppeteerLogin = Ch.PuppeteerLogin
			  AND Po.RealmName = ? 
			  AND Po.PostTimestamp > getInterval_Hour('1 day')
			GROUP BY Po.PuppeteerLogin, Po.PuppetName, Po.RealmName 
			ORDER BY Po.PuppeteerLogin, Po.PuppetName",

		 $realm,
		 $realm
		);

		# Retreive hosts
		$database->GetList(\@hosts, "SELECT p.PuppeteerEmail 
			FROM  Puppeteer p, 
			      RealmSupervisor r 
			WHERE p.PuppeteerLogin=r.PuppeteerLogin 
                          AND r.RealmName=?
			  AND p.PuppeteerSuperPrivs='yes'", $realm);

		# Send mail
		# Initiate mail handle
		$send = tie(*MAIL, 'Ethereal::Mail', $database);

		# Initial settings
		$send->SetSubject("Subject: Statistics for $realm from $start to $end");

		# Recipients
		$send->AddTo(@hosts);


		# Main body
		# Title line
		print MAIL "Statistics for $realm from $start to $end\n\n";

		# Realm statistics
		# Hourly stats
		print MAIL "Realm Statistics\n-------------------------\n\n";
		print MAIL "Time  Posts      Hits  \n";
		print MAIL "----- ---------- ----------\n";
	
		# Cycle through
		for ($count=1; $count < 25; $count++)
		{
			# Search ranges
			$end   = $count - 1;
			$start = $count;

			# Queries
			# Posts
			($lposts) = $database->DataGet("SELECT COUNT(*) 
				FROM Post 
				WHERE RealmName=?
				  AND PostTimestamp > getInterval_Hour('$start hour')
				  AND PostTimestamp < getInterval_Hour('$end hour')
				  AND PostTo <> 'system'
				  AND PostClass NOT IN ('banner','dice','duplicate','system')",

			 $realm
			);

			# Defined
			$lposts  = (defined($lposts)) ? $lposts : 0;
			$gposts += $lposts; 


			# Hits
			($lhits)  = $database->DataGet("SELECT COUNT(*)
				FROM Post
				WHERE RealmName=?
				  AND PostTimestamp > getInterval_Hour('$start hour')
				  AND PostTimestamp < getInterval_Hour('$end hour')",

			 $realm
			);

			# Definition
			$lhits  = (defined($lhits)) ? $lhits : 0;
			$ghits += $lhits;


			# Reuse start
			($start) = $database->DataGet("SELECT getTime(getInterval_Hour('$start hours'))");

			# Display of information
			print MAIL $start . "\t" . FormatNum($lposts, 8) . FormatNum($lhits, 8) . "\n";
		}

		# General Information
		print MAIL "Total\t" . FormatNum($gposts, 8) . FormatNum($ghits, 8) . "\n\n\n";


		# Who was on
		print MAIL "Who was on?\n-------------------------\n\n";

		# Cycle through
		for ($count=1; $count < 25; $count++)
		{
			# Search ranges
			$end   = $count - 1;
			$start = $count;

			# Puppets and puppeteers
			$database->GetList(\@whos, "SELECT Po.PuppetName || ' (' || Po.PuppeteerLogin || ')'
				FROM  Post Po, 
				      Puppet Ch 
				WHERE Po.PuppeteerLogin = Ch.PuppeteerLogin 
				  AND Po.RealmName = ? 
				  AND Po.PostTimestamp > getInterval_Hour('$start hour') 
				  AND Po.PostTimestamp < getInterval_Hour('$end hour') 
				GROUP BY Po.PuppeteerLogin, Po.PuppetName, Po.RealmName 
				ORDER BY Po.RealmName, Po.PuppetName",

			 $realm
			);

			# Reuse start
			($start) = $database->DataGet("SELECT getTime(getInterval_Hour('$start hours'))");

			# Display title
			print MAIL "$start\n-------------------------\n\n";


			# Cycle and display
			while ($who = shift @whos)
			{
				print MAIL "\t$who\n";
			}
		}

		# Userlist
		print MAIL "\n\nUserlist\n-------------------------\n\n";

		# Cycle through
		while ($who = shift(@users))
		{
			print MAIL "\t$who\n";
		}

		# Database addition
	 	$database->Quick("INSERT INTO Statistics
			(RealmName,
			 StatHits,
			 StatPosts)
			VALUES(?,?,?)",

		 $realm,
		 $ghits,
		 $gposts);


		# Simple count
		# Logins
		($pers) = $database->DataGet("SELECT COUNT(DISTINCT(PuppeteerLogin))
			FROM Post
			WHERE RealmName=?
			  AND PostTimestamp > getInterval_Hour('1 day')
			  AND PuppeteerLogin IS NOT NULL", $realm);

		# Handles
		($chars) = $database->DataGet("SELECT COUNT(DISTINCT(PuppetName))
			FROM Post
			WHERE RealmName=?
			  AND PostTimestamp > getInterval_Hour('1 day')
			  AND PuppeteerLogin IS NOT NULL", $realm);

		# List
		print MAIL "\n\nMembers\n";
		print MAIL "--------\n";
		print MAIL "Users   : " . FormatNum($pers, 6)  . "\n";
		print MAIL "Handles : " . FormatNum($chars, 6) . "\n";


		# Post breakup
		# Initial Display
		print MAIL "\n\nCount    Post type\n";
		print MAIL "------- ----------------------------------\n";

		# Prepare
		$database->Pull(\$statement, "SELECT
			  COUNT(*) AS  \"PostCount\",
			  PostClass AS \"PostClass\"
			FROM Post
			WHERE RealmName=?
			  AND PostTimestamp > getInterval_Hour('1 day')
			GROUP BY PostClass
			ORDER BY PostClass",

		 $realm);

		# Find
		while ($res = $statement->fetchrow_hashref())
		{
			# Display all the results
			print MAIL FormatNum($res->{'PostCount'}, 7) . $res->{'PostClass'} . "\n";
		}

		# Finish up
		$statement->finish();


		# Retreive averages
		# Month
		my ($mposts_avg, $mhits_avg, $mposts_sum, $mhits_sum) = $database->DataGet("SELECT
			   ROUND(AVG(StatPosts)),
			   ROUND(AVG(StatHits)),
			   SUM(StatPosts),
			   SUM(StatHits)
			 FROM Statistics
			 WHERE StatTimestamp >= DATE_TRUNC('month', LOCALTIMESTAMP) 
			   AND RealmName=?", $realm);

		# Yearly
		my ($yposts_avg, $yhits_avg, $yposts_sum, $yhits_sum) = $database->DataGet("SELECT
			   ROUND(AVG(StatPosts)),
			   ROUND(AVG(StatHits)),
			   SUM(StatPosts),
			   SUM(StatHits)
			 FROM Statistics
			 WHERE StatTimestamp >= DATE_TRUNC('year', LOCALTIMESTAMP)
			   AND RealmName=?", $realm);


		# Cumulative stats
		print MAIL "\n\nCumulative Statistics\n-------------------------\n\n";
		print MAIL "Time\tPosts\t\tHits\n------- ------- ------- ------- -------\n";
		print MAIL "Month\t"
			. FormatNum($mposts_sum, 7)
			. FormatNum($mposts_avg, 7)
			. FormatNum($mhits_sum, 7)
			. FormatNum($mhits_avg, 7) . "\n";
		print MAIL "Year\t"
			. FormatNum($yposts_sum, 7)
			. FormatNum($yposts_avg, 7)
			. FormatNum($yhits_sum, 7)
			. FormatNum($yhits_avg, 7) . "\n";

		# Close and send
		close(MAIL);


		# Automated archiving
		my ($mail) = $database->DataGetRealmArchive($realm);

		# Only if there is a need 
		if (defined($mail))
		{
			# Create needed parameters
			my %args = (
				GETPOSTS => 0,
				ROOM     => $realm, 
				USER	 => $realm
			 );

			# Call pull
			$pull->AutomatedPosts($database, undef, \%args, $mail);
		}
	}
}


####################
# List Summary
#
# Somewhat inneficient function that will display a summary all the useage
# in all the rooms.   Public or otherwise.
#

sub ListSummary
{
	####################
	# Data members
	my $database = shift;					# Database handle
	my $int      = shift;					# Interval

	my $statement;						# Database statement
	my $cumulative;						# Database results hash

	my $send;						# Mail sender
	my $tstamp;						# Timestamp

	my $number;						# Number
	my $name;						# Name of room

	my $users;						# Amount of users
	my $puppets;						# Amount of puppets
	my $realm;						# Realm name

	my $day;						# Users in day
	my $week;						# Users in week
	my $month;						# Users in month
	my $expire;						# About to expire
	my $elapsed;						# Calculation of elapsed time

	my $time;						# Time to calculate

	my $thits;						# Total hits
	my $tposts;						# Total posts

	my $email;						# Email target
	my @email;						# Email targets

	my %hits;						# Total hits
	my %posts;						# Total posts
	my %system;						# System hash


	####################
	# Program area

	# Ensure defined
	$int =  (defined($int)) ? $int : '-daily';

	# Determin interval
	if ($int =~ /weekly/)
	{ $time = 7;  $int = 'Weekly';}
	elsif ($int =~ /monthly/)
	{ $time = 31; $int = 'Montly';}
	else
	{ $time = 1;  $int = 'Daily'; }


	# Time adjustment
	$time = (defined($time)) ? $time : 1;


	# Retreive hash
	$database->GetHashSystem(\%system);

	# Retreive Email
	$database->GetList(\@email, "SELECT PuppeteerEmail
		FROM  Puppeteer
		WHERE PuppeteerAdminPrivs='yes'
		GROUP By PuppeteerEmail");

	# Pull timestamp
	($tstamp) = $database->DataGetDate();


	# Prepare and execute
	# Total posts
	$database->Pull(\$statement, "SELECT
		  COUNT(*)  AS \"Count\", 
		  RealmName AS \"Realm\"
		FROM  Post
		WHERE PostTimestamp > getInterval_Hour('$time days')
		GROUP BY RealmName
		ORDER BY RealmName");

	while (($number, $name) = $statement->fetchrow())
	{ $hits{$name} = $number };

	# Finish
	$statement->finish();


	# Accurate posting
	$database->Pull(\$statement, "SELECT
		  COUNT(*)  AS \"Count\", 
		  RealmName AS \"Realm\"
		FROM  Post 
		WHERE PostTo <> 'system' 
		  AND PostTimestamp > getInterval_Day('$time days')
		  AND PostClass NOT IN ('banner','dice','duplicate','system')
		GROUP BY RealmName
		ORDER BY RealmName");

	while (($number, $name) = $statement->fetchrow())
	{ $posts{$name} = $number };

	# Finish
	$statement->finish();


	# Set totals
	$thits  = 0;
	$tposts = 0;


	# Initiate Email composition
	# Create instance
	$send = tie(*MAIL, 'Ethereal::Mail', $database);

	# Initial settings
	$send->SetSubject("$int System Statistics ($tstamp)");

	# Recipients
	$send->AddTo(@email);


	# Room useage
	# Initial display
	print MAIL "Daily statistics\n";
	print MAIL "Posts      Hits       Realm Name\n";
	print MAIL "---------- ---------- ----------------------------------\n";

	# Retreive and display
	foreach $name (sort(keys(%hits)))
	{
		# Ensure integrity
		$posts{$name} = (defined($posts{$name})) ? $posts{$name} : 0;
 
		# Display to screen
		print MAIL FormatNum($posts{$name}) . FormatNum($hits{$name}) . $name . "\n"; 

		# Add to totals
		$thits  += $hits{$name};
		$tposts += $posts{$name};
	}


	# Insert totals
	$database->Quick("INSERT INTO Statistics
		(StatHits,
		 StatPosts)
		VALUES (?,?)",

	 $thits,
	 $tposts);
	

	# Cumulative room useage
	# Display
	print MAIL "\n\nCumulative Statistics\n";
	print MAIL "Posts                 Hits            Realm Name\n";
	print MAIL "---------- ---------- ---------- ---------- ----------------------------------\n";

	# Fetch
	$database->Pull(\$statement, "SELECT
		  RealmName             AS \"RealmName\",
	  	  ROUND(AVG(StatHits))  AS \"StatHitsAvg\",
		  ROUND(AVG(StatPosts)) AS \"StatPostsAvg\",
		  SUM(StatPosts)        AS \"StatPostsSum\",
		  SUM(StatHits)         AS \"StatHitsSum\"
		FROM Statistics
		WHERE StatTimestamp >= DATE_TRUNC('month', LOCALTIMESTAMP)
		GROUP BY RealmName
		ORDER BY RealmName");

	# Cycle through
	while ($cumulative = $statement->fetchrow_hashref())
	{
		# Display to screen
		print MAIL FormatNum($cumulative->{'StatPostsSum'})
	         	. FormatNum($cumulative->{'StatPostsAvg'})
			. FormatNum($cumulative->{'StatHitsSum'})
			. FormatNum($cumulative->{'StatHitsAvg'})
			. $cumulative->{'RealmName'} . "\n"; 
	}
	$statement->finish();
			

	# Chat Preferences
	# Initial Display
	print MAIL "\n\nHits    Chat Mode\n";
	print MAIL "------- ----------------------------------\n";

	# Prepare and execute
	$database->Pull(\$statement, "SELECT COUNT(*), 
		  PuppeteerChatPref 
		FROM Puppeteer 
		GROUP BY PuppeteerChatPref");

	# Retreive and display
	while (($number, $name) = $statement->fetchrow())
	{ print MAIL FormatNum($number, 7) . $name . "\n"; }

	# Finish
	$statement->finish();


	# Formatting
	# Initial Display
	print MAIL "\n\nHits    Formatting\n";
	print MAIL "------- ----------------------------------\n";

	# Prepare and execute
	$database->Pull(\$statement, "SELECT COUNT(*), 
		  PuppeteerFormatting
		FROM Puppeteer 
		GROUP BY PuppeteerFormatting");

	# Retreive and display
	while (($number, $name) = $statement->fetchrow())
	{ print MAIL FormatNum($number, 7) . $name . "\n"; }

	# Finish
	$statement->finish();


	# Javascript
	# Initial Display
	print MAIL "\n\nHits    Javascripting\n";
	print MAIL "------- ----------------------------------\n";

	# Prepare and execute
	$database->Pull(\$statement, "SELECT COUNT(*), 
		  PuppeteerJavascript
		FROM Puppeteer 
		GROUP BY PuppeteerJavascript");

	# Retreive and display
	while (($number, $name) = $statement->fetchrow())
	{ print MAIL FormatNum($number, 7) . $name . "\n"; }

	# Finish
	$statement->finish();


	# Puppet Useage
	# Initial display
	print MAIL "\n\nFreqency Count   Percent\n";
	print MAIL "-------- ------- -------\n";

	# Totals
	($users)   =  $database->DataGet('SELECT COUNT(*) FROM Puppeteer WHERE PuppeteerBan=?', 'no');
	($puppets) =  $database->DataGet('SELECT COUNT(*) FROM Puppet');

	# Day use
	($day)     = $database->DataGetFrenquency(1);
	printf MAIL "Day      %7i   %5.2f\n", $day, (($day/$users)*100);	

	# Weekly use
	($week)    = $database->DataGetFrenquency(7);
	printf MAIL "Week     %7i   %5.2f\n", $week, (($week/$users)*100);	

	# Monthly use
	($month)   = $database->DataGetFrenquency(30);
	printf MAIL "Month    %7i   %5.2f\n", $month, (($month/$users)*100);	

	# About to expire
	($expire)  = $database->DataGet("SELECT COUNT(*) 
		FROM Puppeteer 
		WHERE PuppeteerTimestamp < getInterval('89 days')");
	printf MAIL "Expire   %7i   %5.2f\n", $expire, (($expire/$users)*100);	

	print  MAIL "\n\nTotals\n-------\nPuppeteers: " . format_number($users, 0, 0) . "\nPuppets: " . format_number($puppets, 0, 0) . "\n";


	# Close and send
	close(MAIL);
}


####################
# Expire Audits
# 
# Will expire all audit entries older then a certain time, this
# will however not expire banned user information.

sub ExpireAudit
{
	####################
	# Data members
	my $database = shift;					# Database handle


	####################
	# Program area

	# Evaluate
	eval
	{
		# Remove audit
		$database->Write("DELETE FROM Audit
			WHERE Audit.AuditTimestamp < getInterval('45 days')
			  AND NOT EXISTS (SELECT Puppeteer.PuppeteerLogin
				FROM Puppeteer
				WHERE Puppeteer.PuppeteerBan='yes'
				  AND Puppeteer.PuppeteerLogin=Audit.PuppeteerLogin)");

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Audit");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}


####################
# Expire Bans
# 
# Will expire all bans older then a certain time.

sub ExpireBans
{
	####################
	# Data members
	my $database = shift;					# Database handle


	####################
	# Program area

	# Evaluate block
	eval
	{
		# Remove old posts
		$database->Write("DELETE FROM PuppeteerRestriction
			WHERE RestrictionTimestamp < LOCALTIMESTAMP");

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE PuppeteerRestriction");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}

####################
# Expire mail
# 
# Will expire all mail older then a certain time.

sub ExpireMail
{
	####################
	# Data members
	my $database = shift;					# Database handle

	my %system;						# System hash


	####################
	# Program area

	# Calculate time
	$database->GetHashSystem(\%system);

	# Evaluate
	eval
	{
		# Remove old posts
		$database->Write("DELETE FROM Mail 
			WHERE MailTimestamp < getInterval('$system{SetTimeoutMail} days')");

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Mail");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}

####################
# Expire Offline
# 
# Will expire all offline posts of the appropriate age

sub ExpireOffline
{
	####################
	# Data members
	my $database = shift;					# Database handle

	my %system;						# System hash


	####################
	# Program area

	# Calculate time for which older are destroyed
	$database->GetHashSystem(\%system);

	# Evaluate block
	eval
	{
		# Cleanup
		# Remove old posts
		$database->Write("DELETE FROM Offline
			WHERE OfflineTimestamp < getInterval('$system{SetTimeoutOffline} days')");

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Offline");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}


####################
# Expire Puppets
# 
# Expires inactive puppets this will occur at different intervals then the 
# expiration of the puppeteer

sub ExpirePuppet
{
	####################
	# Data members
	my $database = shift;					# Database handle

	my %system;						# System hash


	####################
	# Program area

	# Calculate time for which older are destroyed
	$database->GetHashSystem(\%system);

	# Evaluate
	eval
	{
		# Remove old posts
		$database->Write("DELETE FROM Puppet 
			WHERE PuppetActive IS NOT NULL
			  AND PuppetActive < getInterval('$system{SetTimeoutPuppet} days')");

		# Commit
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Puppet");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}

####################
# Expire Puppeteer
#
# Will remove all puppeteers that have lapsed on their membership times.  Which
# is determine in the system DBM.

sub ExpirePuppeteer
{
	####################
	# Data members
	my $database = shift;					# Database handle

	my $page;						# Home page handler

	my $home;						# Home page
	my $gallery;						# Gallery name
	my $puppeteer;						# Individual puppeteer
	
	my @galleries;						# List of galleries
	my @puppeteers;						# List of puppeteers

	my %contrib;						# Home page hash
	my %system;						# System hash


	####################
	# Program area
	
	# Link with hash
	$database->GetHashContrib(\%contrib);
	$database->GetHashSystem(\%system);

	# Create instance
	$page = new Ethereal::Home($database);


	# Rerieve list of puppeteers
	$database->GetList(\@puppeteers, "SELECT PuppeteerLogin 
		FROM Puppeteer 
		WHERE PuppeteerTimestamp < getInterval('$system{SetTimeoutPuppeteer} days') 
		  AND PuppeteerAdminPrivs=?
		  AND PuppeteerSuperPrivs=?
		  AND PuppeteerBan=?
		  AND PuppeteerVerified=?",
	 'no',
	 'no',
	 'no',
	 'no');

	# Evaluate block
	eval
	{
		# Run though list of puppeteers	
		foreach $puppeteer (@puppeteers)
		{
			# Home page handling
			# Retreive home
			($home) = $database->DataGetPuppeteerHome($puppeteer);
	
			# Remove folder
			if (defined($home))
		 	{
				# Attempt to remove, else error 
				unless ($page->DirClear($contrib{'SetHomeRoot'} . '/' . $home)) 
				{
					# Necessary
					die("ERROR: Home directory " . $contrib{'SetHomeRoot'} . '/' . $home . " not cleared ($!)");
				}
			}

			# Gallery Handling
			# Retreive list of galleries
			$database->GetList(\@galleries, "SELECT GalleryID
				FROM Gallery
				WHERE PuppeteerLogin=?", $puppeteer);
		
			# Run though list of puppets
			while ($gallery = shift @galleries)
			{
				# Attempt to remove, else error 
				unless ($page->DirClear("$system{SetBaseDir}/$gallery")) 
				{
					# Necessary for operation
					die("ERROR: Gallery $system{SetBaseDir}/$gallery not cleared ($!)");
				}			
			}

			# Remove all instances of the puppeteers
			$database->Write("DELETE FROM Puppeteer WHERE PuppeteerLogin=?", $puppeteer);
		}

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Puppeteer");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}


####################
# Expire Realm
#
# Will remove all private realms after elapsed time.

sub ExpireRealm
{
	####################
	# Data members
	my $database = shift;					# Database handle
	
	my %system;						# System hash


	####################
	# Program area

	# Calculate time
	$database->GetHashSystem(\%system);

	# Evaluate
	eval
	{
		# Clear out
		$database->Write("DELETE FROM Realm
			WHERE Realm.RealmName IN (SELECT Realm.RealmName
				FROM  Realm
				WHERE Realm.RealmPublic='no'
				AND Realm.RealmTimestamp < getInterval('$system{SetTimeoutRealm} days')
				AND NOT EXISTS (SELECT DISTINCT Post.RealmName
					FROM Post
					WHERE Post.RealmName=Realm.RealmName)
				AND NOT EXISTS (SELECT DISTINCT Offline.RealmName
					FROM Offline
					WHERE Offline.RealmName=Realm.RealmName))");

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Realm");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}


####################
# Expire Registrations
# 
# Will expire all registration older then a certain taime who have not completed
# their registration.

sub ExpireRegistrations
{
	####################
	# Data members
	my $database = shift;					# Database handle
	my $statement;						# Database statement

	my $time;						# Stale time
	my $puppeteer;						# Individual puppeteer

	my %system;						# System hash


	####################
	# Program area

	# Calculate time
	$database->GetHashSystem(\%system);

	# Evaluate block
	eval
	{
		# Clear expired registrations
		$database->Write("DELETE FROM Puppeteer
			WHERE PuppeteerLogin IN (SELECT PuppeteerLogin
				FROM  PuppeteerRegistration
				WHERE PuppeteerTimeStamp < getInterval('$system{SetTimeoutRegister} days'))");

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Puppeteer");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}

####################
# Update Forcast
#
# Updates the forcasts for all of the realms that are making use of this system
# allowing the hosts to get fresh weather information without a long wait.

sub ForcastUpdate
{
	####################
	# Data Members
	my $database = shift;					# Database handle

	my $forc;						# Handle to weather

	my $station;						# Individual station
	my @stations;						# List of stations
	my %stations;						# Stations and forcasts
	

	####################
	# Program Area

	# Retreive list
	$database->GetList(\@stations, "SELECT RealmICAO
		FROM  Realm
		WHERE RealmICAO IS NOT NULL
		GROUP BY RealmICAO");


	# Initialize handle
	$forc = new Ethereal::Weather();

	# Retreive forcasts
	%stations = $forc->Get(@stations);


	# Evaluate for speed
	eval
	{
		# Cycle and update database
		foreach $station (keys(%stations))
		{
			# Update with new information
			$database->Write("UPDATE Realm
				SET   RealmMETAR=?
				WHERE RealmICAO=?",

			 $stations{$station},
			 $station);
		}

		# Commit
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");

		# Undo incomplete changes
		$database->Rollback();
	}
}

####################
# Print Banner
#
# Will look to see which realms are actually busy, then using that
# information will print a predefined banner to all of those realms
# publicly.   Use WebBanner to change the default template to 
# display.

sub PrintBanner
{

	####################
	# Data Members
	my $database = shift;					# Database handle
	my $type     = shift;					# How to handle things

	my $rotate;						# Rotation handle
	my $statement;						# Database statement

        my $banner;						# Banner to display
	my $line;						# Captured line
	my $realm;						# Current realm

	my $class  = 'banner';					# Posting class
	my $rights = 'display';					# Display setting
	my $target = 'all';					# Designated target

	my %realm;						# Realm handler
	my %system;						# System handler


	####################
	# Program Area

	# Populate hash
	$database->GetHashRealm(\%realm);
	$database->GetHashSystem(\%system);

	# Ensure definition
	$type = (defined($type)) ? $type : 'database';


        # Pull banner
	if ($type eq 'database')
	{
		# Create handle
		$rotate = new Ethereal::Rotate(
			DB    => $database,
			ZONE  => 'banner',
			STYLE => 'time'
		 );

		# Fetch a banner
		$banner = $rotate->Pull();
	}

	else
	{
		# Safety assign
		$banner = '';

		# Make use of STDIN for input
		while ($line = <STDIN>)
		{
			# Capture and append
			$banner = $banner . $line;
		}
	}

	# Data write
	if ((defined($banner))
	 && (length($banner) > 15))
	{
		# Determine rooms
		# Prepare and execute
		$database->Pull(\$statement, "SELECT Po.RealmName
			FROM Post Po, Realm Re
			WHERE Po.RealmName = Re.RealmName
			  AND Re.RealmBanner=?
			  AND Po.PostTimestamp > getInterval('10 Minutes')
	                  AND Po.PuppetName IS NOT NULL
			GROUP BY Po.RealmName", $rights);

		# Evaluate for speed
		eval
		{
			# Make posts
			while (($realm) = $statement->fetchrow())
			{
				# Make post
				$database->Write("INSERT INTO Post
					(RealmName,
					 PostTo,
					 PostClass,
					 PostFullFormat,
					 PostImagelessFormat,
					 PostPartialFormat)
					VALUES (?,?,?,?,?,?)",

				 $realm,
				 $target,
				 $class,
				 $banner,
				 $banner,
				 $banner
				);
			}

			# Commit changes
			$database->Commit();
		};

		# If block failed
		if ($@)
		{
			# Issue warning
			warn("Transaction aborted: $@");
	
			# Undo incomplete changes
			$database->Rollback();
		}

		# Finish query
		$statement->finish();
	}
}


####################
# Print Stream
#
# This function will be used to post a notice to the realms, in which the
# user will be notified of a reset.    This will also send off appropriate
# information as to kill the users stream entirely. 

sub PrintStream
{

	####################
	# Data Members
	my $database = shift;					# Database handle
	my $mes      = shift;					# Message
	my $kill     = shift;					# Kill message

	my $statement;						# Database statement

        my $banner;						# Banner to display
	my $class    = 'banner';				# Posting class
	my $command  = 'stream';				# Stream kill command

	my $user;						# Username
	my $realm;						# Current realm

	my %realm;						# Realm handler


	####################
	# Program Area

	# Populate hash
	$database->GetHashRealm(\%realm);


	# Message handling
	$mes     = (defined($mes)) ? $mes : $realm{'TxtEndStream'};

	# Final decision on kill
	$command = (!defined($kill)) ? $command : 'none';


        # Pull banner
	$banner  = (!defined($kill)) ? $database->DocumentGetStream() : $database->DocumentGetNotice();

	# Search and replace
	$banner =~ s/MDESCR/$mes/gs;



	# Determine rooms
	# Prepare and execute
	$database->Pull(\$statement, "SELECT DISTINCT ON (PuppeteerLogin)
		  PuppeteerLogin AS \"PuppeteerLogin\",
		  RealmName      AS \"RealmName\"
		FROM Post
		WHERE PostTimestamp > getInterval('10 Minutes')
		  AND PuppetName IS NOT NULL
		ORDER BY PuppeteerLogin, RealmName");

	# Evaluate for speed
	eval
	{
		# Make posts
		while (($user, $realm) = $statement->fetchrow())
		{
			# Make post
			$database->Write("INSERT INTO Post
				(RealmName,
				 PuppeteerLogin,
				 PostTo,
				 PostClass,
				 PostCmd,
				 PostFullFormat,
				 PostImagelessFormat,
				 PostPartialFormat)
				VALUES (?,?,?,?,?,?,?,?)",
	
			 $realm,
			 $user,
			 $user,
			 $class,
			 $command,
			 $banner,
			 $banner,
			 $banner
			);
		}

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Finish query
	$statement->finish();
}

# Variant of the above
sub PrintNotice
{ PrintStream(shift, shift, 0); }


####################
# Purge Param
#
# With the addition of database stored parameters, it has become
# necessary to create a purge utility as to clean things up and
# keep things fast.

sub PurgeParam
{

	####################
	# Data Members
	my $database = shift;					# Database handle

	my $access;						# Last access time
	my $file;						# File list

	my @files;						# Local hashes stored

	####################
	# Program area

	# Cleanup
	eval
	{
		# Remove rows
		$database->Write("DELETE FROM Param
			WHERE ParamTimestamp < getInterval('120 minutes')");

		# Commit
		$database->Commit();


		# Database files
		if (-d $database->{'DBAUTH'})
		{
			# Get list
			@files = glob($database->{'DBAUTH'} . '/*');

			# Listing
			foreach $file (@files)
			{
				# This exists
				if (-f $file)
				{
					# Access time
					$access = -A $file;	

					# Modify appropriatly
					$access = ($access * 86400) / 60;

					# Remove if old
					unlink($file) if ($access > 120);
				}
			}
		}
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Param");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}


####################
# Purge Posts
# 
# Will expire all posts which are older then twenty minutes.

sub PurgePosts
{
	####################
	# Data members
	my $database = shift;					# Database handle
	my $time     = shift;					# Stale time

	####################
	# Program Area

	# Time adjustment
	$time = (defined($time)) ? $time : 1;

	# Evaluate block
	eval
	{
		# Cleanup
		# Remove old posts
		$database->Write("DELETE FROM Post 
			WHERE PostTimestamp < getInterval('$time days')");

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Post");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}


####################
# Reset Statistics
#
# Will reset cumulative statistics, essentially clearing the values within
# the table itself.   Good for months or quarterly resets on statistics.

sub ResetStats
{
	####################
	# Data Members
	my $database = shift;					# Database handle

	####################
	# Program Area

	# Evaluate block
	eval
	{
		# Complete database operation
		$database->Write("DELETE FROM Statistics
			WHERE StatTimestamp < DATE_TRUNC('year', LOCALTIMESTAMP)");

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");
	
		# Undo incomplete changes
		$database->Rollback();
	}

	# Successful
	else
	{
		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Vacuum
		$database->Write("VACUUM ANALYSE Statistics");

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}

####################
# Vacuum
#
# Will take care of the annoyance of vaccuming tables, which is a requirement
# for tables under PostgreSQL.

sub Vacuum
{
	####################
	# Data Members
	my $database = shift;					# Database handle
	my $type     = shift;					# Decisions

	my $search;						# Seach path
	my $table;						# Single table
	my @tables;						# Table list

	my %system;						# Realm handler

	####################
	# Program Area

	# Ensure definition
	$type = (defined($type)) ? $type : 'fast';


	# Collect search path
	($search) = $database->DataGet("SHOW search_path");

	# Split appart
	$search =~ s/^(\w+)\,\s(\w+)$/$1/i;

	# Collect tables
	$database->GetList(\@tables, "SELECT relname
		FROM pg_catalog.pg_stat_all_tables
		WHERE schemaname=?
		ORDER BY schemaname", $search);


	# Commit to prevent screwups
	$database->Commit();


	# Fast maintenance
	# Recommended for daily use
	if ($type eq 'fast')
	{
		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Loop and clean
		foreach $table (@tables)
		{
			# Vacuum
			$database->Write("VACUUM ANALYSE $table");
		}

		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}

	# Verbose mode
	elsif ($type eq 'verbose')
	{

		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Loop and clean
		foreach $table (@tables)
		{
			# Readability
			print "TABLE: $search.$table\n";

			# Vacuum
			$database->Write("VACUUM ANALYSE VERBOSE $table");

			# Readability
			print "\n";
		}

		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}

	# Complete locking maintenance
	# Recommended for monthly use
	if ($type eq 'complete')
	{
		# Times
		my $start;					# Time start time
		my $stop;					# Stop time

		# Warning
		# Populate hash
		$database->GetHashSystem(\%system);


		# Disable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 1;

		# Standard maintenance
		print "$search\n";

		# Loop and clean
		foreach $table (sort(@tables))
		{
			# Start time
			$start = time;

			# Informational
			print "  " . ucfirst($table) . " ";


			# Vacuum
			$database->Write("VACUUM FULL ANALYSE $table");
			$database->Write("REINDEX TABLE $table");


			# End time
			$stop = time-$start;

			# Cleanup
			print "(" . $stop . " seconds)\n";
		}

		# Spacer
		print "\n";

		# Normal vaccum to make sure
		Vacuum($database);

		# Enable autocommit
		$database->{'HANDLE'}->{'AutoCommit'} = 0;
	}
}
